/**

    Copyright 2007 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package it.eng.spagosoa.smx.listeners;

import java.io.Serializable;

import javax.jbi.messaging.MessageExchange;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.ObjectMessage;
import javax.jms.Queue;
import javax.jms.QueueConnection;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueSender;
import javax.jms.QueueSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.jbi.event.ExchangeEvent;
import org.apache.servicemix.jbi.event.ExchangeListener;

public class AuditingExchangeASyncListener implements ExchangeListener {

	private static final Log log = LogFactory.getLog(AuditingExchangeASyncListener.class);
	
	private ConnectionFactory connectionFactory = null;
	private Destination destinationQueue = null;
	
	public AuditingExchangeASyncListener() {
		super();
	}
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}
	
	public ConnectionFactory getConnectionFactory() {
		return connectionFactory;
	}

	public void setConnectionFactory(ConnectionFactory connectionFactory) {
		this.connectionFactory = connectionFactory;
	}

	public Destination getDestinationQueue() {
		return destinationQueue;
	}

	public void setDestinationQueue(Destination destinationQueue) {
		this.destinationQueue = destinationQueue;
	}

	
	public void exchangeSent(ExchangeEvent event) {
		 MessageExchange exchange = event.getExchange();
		 
		 QueueConnection queueConnection = null;
		 QueueSession queueSession = null;
		 QueueSender queueSender = null;
    	
    	try {
    		debug("Creating Queue Connection.....");
        	queueConnection =
        		((QueueConnectionFactory)getConnectionFactory()).createQueueConnection();
        	
        	queueConnection.start();
        	debug("Queue Connection is OK....");
        	debug("Creating Queue Session.......");
        	queueSession =
        		queueConnection.createQueueSession(false, javax.jms.Session.AUTO_ACKNOWLEDGE);
        	
        	debug("Queue Session is OK.....");
        	debug("Creating Queue Sender....");
        	queueSender = queueSession.createSender((Queue)destinationQueue);
        	queueSender.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
        	debug("Creating Queue Sender is OK....");
        	
        	debug("Creating Object Message ");
        	ObjectMessage message = queueSession.createObjectMessage();
        	debug("Setting exchange in object message....");
        	message.setObject((Serializable)exchange);
        	debug("Setted exchange in object message....");
        	
        	debug("Sending in queue............");
        	queueSender.send(message);
        	debug("Send OK ");
    	}
    	catch (Exception ex) {
    		log.error(ex);
    	}
    	finally {
    		try {
        		if (queueSender != null) {
                	queueSender.close();
        		}
        		if (queueSession != null) {
                	queueSession.close();
        		}
        		if (queueConnection != null) {
                	queueConnection.close();
        		}
    		}
    		catch (Exception ex) {
    			log.error(ex);
    		}
    	}
    	
	}

	public void exchangeAccepted(ExchangeEvent event) {
		// Nothing to do
	}

}
